local debug=function(arg) if type(arg)=='table' then print(serpent.block(arg)) else print(arg) end end
local on_replacement = script.generate_event_name()--the event that will be raised when the entity is destroyed and another created
--kind of useless actually, since is not universally accepted
remote.add_interface('liftoff', {
	on_replacement=function() return on_replacement end,
	info=function()
		print('liftoff speeds')
		for k,v in pairs(global.def) do 
			game.player.print(k..' speed: '..v.v..' runway lenght: '..v.s)
		end
	end,
})

require 'planes'
script.on_init(function() init() end)
script.on_configuration_changed(function() init() end)

function init()
	global.planes=global.planes or {--[[
		[i]={
		[1]=entity,
		[2]=grounded,
		[3]=float(runway_covered),
		[4]=int(heigth)
		}
	]]}--table to keep track of vehicles capable of flying
	global.pilots=global.pilots or {--[[
		[player.index]={
			veh=entity,
			speed=gui_handle, 
			height=gui_handle,
			data=reference_to_global_planes,
			}
	]]}--a table to keep track of players that are inside of vehicles, should be always updated with almost freshest reference to plane
	global.falling=global.falling or {--[[
		[player.index]={
			[1]=velocity,
			[2]=direction_cos
			[3]=direction_sin
		}
	]]}--keeps track of velocity at which player is flinged if out of plane at full speed
	global.def={--[[
		[entity.name]={
			grounded=grounded_dummy_entity.name,
			v=liftoff_speed,
			d=runway_length,
			shadow=shadow_entity.name,
		}
	]]}--values often needed for updates
	global.fed={--[[
		[grounded_entity.name]=entity.name
	]]}
	global.edf={--[[
		[shadow_entity.name]=entity.name
	]]}
	for _,p in pairs(game.entity_prototypes) do
		if p.type=='car' and string.find(p.name,"#grounded#") then
			--local _,_,speed,speed2,name=string.find(p.name,"#grounded#(%d+)#(%d+)#(.+)")--thi is not used 
			local _,_,name=string.find(p.name,"#grounded#(.+)")
			global.def[name]={
				grounded=p.name,
				shadow="#shadow#"..name,
				v=planes_definition[name].lift_velocity,
				s=planes_definition[name].runway_length, 
				}
		end
	end
	for k,v in pairs(global.def) do
		global.fed[v.grounded]=k
		global.edf[v.shadow]=k
	end
end

function place(e,new_name)
	g=e.surface.create_entity{name=new_name,position=e.position,force='neutral'}
	g.insert("dummy-plane-fuel")
	g.orientation=e.orientation
	g.speed=e.speed
	g.minable=false
	g.operable=false
	g.riding_state=e.riding_state
	return g
end


function onbuilt(ev)
	local e=ev.created_entity
	if global.def[e.name] then
		if e.surface.can_place_entity{name="car",position=e.position} then
			local d=global.def[e.name]
			g=place(e,d.grounded)
			g.force=e.force
			table.insert(global.planes,{e,g,global.def[e.name].s,0})
			--debug('built')
			--debug(global.def[e.name].s)
		else
			game.players[ev.player_index].insert(e.name)
			e.destroy()
		end
	end
end
script.on_event(defines.events.on_built_entity,onbuilt)--car cannot be built by robots

function find(t,field,e)
	for i=1,#t do if t[i][field]==e then return i end end
end
script.on_event(defines.events.on_entity_died,function(ev)
	local e=ev.entity
	if global.def[e.name] or global.fed[e.name] then
		local ind=(global.def[e.name] and 1) or 2
		local g=global.planes
		local j=find(g,ind,e)
		if j then
			g[j][3-ind].die()
			table.remove(g,j) 
		end
	end
end)
script.on_event({defines.events.on_robot_pre_mined_item,defines.events.on_preplayer_mined_item,},function(ev)
	local e=ev.entity
	if global.def[e.name] then
		local g=global.planes
		local j=find(g,1,e)
		if not j then return end
		g[j][2].destroy()
		table.remove(g,j) 
	end
end)

script.on_event(defines.events.on_player_driving_changed_state,
function(ev)
	local pindex=ev.player_index
	local p=game.players[ev.player_index]
	if p.vehicle then
		local name=p.vehicle.name
		if global.fed[name] then
			local j=find(global.planes,2,p.vehicle)
			global.planes[j][1].passenger=p.character
			game.raise_event(defines.events.on_player_driving_changed_state,ev)
		end
		if global.def[name] then 
			--debug('guing')
			if not find(global.planes,1,p.vehicle) then --if plane was built before installing mod
				onbuilt({created_entity=p.vehicle})
			end
			local gui=p.gui.top.add{type='frame',direction='vertical',name='liftoff'}
			local spmeter=gui.add{type='flow',name='speedometer',direction='horizontal'}
			spmeter.add{type='label',name='text', caption={'speed'}}
			local spv=spmeter.add{type='progressbar',size=1}
			local altmeter=gui.add{type='flow',name='altmeter',direction='horizontal'}
			altmeter.add{type='label',name='text', caption={'height'}}
			local alv=altmeter.add{type='progressbar',size=1}
			global.pilots[pindex]={
				veh=p.vehicle,
				speed=spv,
				height=alv,
				data=global.planes[find(global.planes,1,p.vehicle)]
				}
		elseif global.edf[p.vehicle.name] then
			p.vehicle.passenger=nil
			game.raise_event(defines.events.on_player_driving_changed_state,ev)
		end
	elseif global.pilots[pindex] then
		local veh= global.pilots[pindex].veh
		global.falling[pindex]={veh.speed,math.sin(veh.orientation*math.pi*2),math.cos(veh.orientation*math.pi*2)}
		p.gui.top.liftoff.destroy()
		global.pilots[pindex]=nil
	end
end)

script.on_event(defines.events.on_player_left_game,
function(ev)
	local pindex=ev.player_index
	--global.pilots[pindex]=nil
	global.falling[pindex]=nil
end)

local floor= math.floor
local round=function (num)
	return floor(num+0.5)
end
local accelerating=defines.riding.acceleration.accelerating

script.on_event(defines.events.on_tick,
function(ev)
	--update planes
	local g=global.planes
	local def=global.def
	for i=1,#g do
		local d=g[i]
		local e=d[1]
		local g=d[2]
		local par=def[e.name]
		--local d4=d[4]
		--grounded behavior
		if d[4]==0 then
			--collision check
			local ep, gp = e.position, g.position
			if (ep.x~=gp.x) or (ep.y~=gp.y) then
				--die in water
				if e.surface.get_tile(ep.x,ep.y).collides_with("water-tile") then
					if e.passenger then
						p=e.passenger
						e.passenger=nil
						p.die()
					end
					e.speed=1
					e.die()
				elseif g.energy==0 or e.energy==0 then
					--if dummy fuel has expired
					g.insert('dummy-plane-fuel')
					g.teleport(e.position)
				else
					--collision
					e.teleport(g.position)
					e.health=g.health
					e.speed=g.speed
				end
			--another if is merged to the above, it is not particularly related, but it is needed to evade invalid entity checks
			elseif e.speed>par.v then--begin takeoff
				--debug('pulling')
				d[3]=d[3]-e.speed
				if d[3]<0 then
					--debug('pulled')
					--take off
					g.destroy()
					g=place(e,par.shadow)
					d[2]=g
					g.destructible=false
					d[4]=2
				end
			elseif d[3]<par.s then--too slow, inertia
				--debug('soslow')
				d[3]=d[3]+par.v
				e.speed=par.v*.9
				g.speed=e.speed
			else --d[3]>par.s and e.speed<par.v --plane can stop now
				--debug("regain control")
				d[3]=par.s
			end
		elseif d[4]<360 then
			--perform altitude change
			if e.speed>par.v then
				g.teleport({e.position.x+0.005*d[4],e.position.y+.01*d[4]})
				d[4]=d[4]+1
			elseif e.riding_state.acceleration~=accelerating then
				d[4]=d[4]-1
				--debug('going down')
				e.speed=par.v*.9
				g.speed=e.speed
				g.teleport({e.position.x+0.005*d[4],e.position.y+.01*d[4]})
			end
			if e.energy==0 then g.teleport({e.position.x+0.005*d[4],e.position.y+.01*d[4]}) end
			--landing
			if d[4]==1 then
				g.destroy()
				g=place(e,par.grounded)
				g.force=e.force
				d[2]=g
				d[4]=0
			end
		else
			--keep high
			if e.speed<par.v then
				d[4]=d[4]-1
			end
		end
		if e.valid and g.valid then
			--make shadow follow
			g.riding_state=e.riding_state
			g.health=e.health
			g.orientation=e.orientation
		end
	end--for i=1,#g do
	--update falling players
	local f=global.falling
	if ev.tick%10==0 and next(f)  then
		for pindex,tuple in pairs(f) do
			local p=game.players[pindex]
			if (p.valid and p.connected and p.character and p.character.valid) then 
				local v=tuple[1]
				local sin,cos=tuple[2],tuple[3]
				local x,y=p.position.x, p.position.y
				local pos={x+sin*v*10,y+cos*v*10}
				if p.surface.can_place_entity{name="player",position=pos} then p.teleport(pos) end
				p.character.damage(v*v*10,'neutral','impact')
				if ev.tick%100==0 then p.character.damage(v*v*1000,'neutral','impact') end
				tuple[1]=v*.9
				if v<0.01 then f[pindex]=nil end
			end
		end
	end
	--update pilot gui
	if ev.tick%30==0 then
		--debug('gui')
		--debug(global.pilots)
		for i,pilot in pairs(global.pilots) do
			local p=game.players[i]
			--debug(p.index)
			--debug(p.valid)
			--debug(p.connected)
			if p.valid and p.connected then
				if p.vehicle then 
					--debug('updating')
					--debug(p.vehicle.speed/def[p.vehicle.name].v)
					pilot.speed.value=p.vehicle.speed/def[p.vehicle.name].v
					pilot.height.value=pilot.data[4]/360
				else
					p.gui.top.liftoff.destroy()
					global.pilots[p.index]=nil
				end
			end
		end
	end
end)
