--this file contains functions used to generate the entities needed for flight imitation
liftoff={--[[
	to_watt=to_watt
	max_speed=max_speed
	lift_equation=lift_equation
	generate=generate
]]}

local generate=function(plane,shadow_animation)
	--[[takes prototype definition of a plane, i.e. an entity that is flying 
	--generates dummy entities that are used to make that plane grounded and flying
	--the entities are handled 
	--if client has shadow sprites as separate file the table of standard format should be passed as second argument, i.e. variants of:
	----shadow={
      layers ={--you get the idea
		}
      }
	]]
	local grounded=table.deepcopy(plane)
	grounded.render_layer ="lower-object"
	grounded.final_render_layer ="lower-object"
	grounded.collision_box=grounded.collision_box or grounded.selection_box
	grounded.collision_mask=nil
	grounded.resistances=grounded.resistances or {}
	table.insert(grounded.resistances,{type='impact', percent=0,decrease=0})
	grounded.stop_trigger=nil
	grounded.vehicle_impact_sound=nil
	grounded.working_sound=nil
	grounded.open_sound=nil
	grounded.close_sound=nil
	grounded.dying_explosion=nil
	grounded.flags={"pushable","placeable-off-grid"}
	grounded.selection_box={{0,0},{0,0}}
	grounded.terrain_friction_modifier = nil
	--local int,frac= math.modf(math.floor(liftoff.lift_equation(grounded)*100)/100)
	--frac=frac*100
	--grounded.name="#grounded#"..int.."#"..frac.."#"..plane.name
	grounded.name="#grounded#"..plane.name
	
	local shadow=table.deepcopy(grounded)
	shadow.name="#shadow#"..plane.name
	shadow.render_layer="lower-object"
	shadow.terrain_friction_modifier = 0
	shadow.collision_mask={}
	--shadow.consumption="1W"
	if shadow_animation then
		shadow.animation=shadow_animation
	else
		sa=shadow.animation
		if sa.layers then
			for _,l in pairs(sa.layers) do
				l.tint={r=0,g=0,b=0,a=.7}
				l.draw_as_shadow = true
			end
		else
			sa.tint={r=0,g=0,b=0,a=.7}
			sa.draw_as_shadow = true
		end
	end
	data:extend({grounded,shadow})
end

--currently functions below are not used as I've failed to predict the speed of aircraft
--instead the config file `planes.lua` is used to define the lift velocity and runway length

predefined={ --I've failed at automatic generation of theses
	}
local import_planes=function(data) predefined=data end
local lift_equation=function(plane)
	--uses the lift equation to determine the speed of takeoff of a plane
	--lift equation:
	--liftForce=1/2 * airDensity * speed^2 * wingArea * wingCoeficient
	--hence:
	--v=sqrt( weight * someCoeficient / wingArea)
	--the coeficient will be chosen to condition basic gunship from Aircraft mod
	--the wingArea will be based of ingame sprite size of a vehicle
	local res=predefined[plane.name]
	if not res then
		local weight=plane.weight 
		local area=1
		if plane.animation.layers then
			for _,l in pairs(layers) do
				local a=l.width*l.height*(l.scale or 1)^2
				area= (area>a) and area or a
			end
		else
			local an=plane.animation
			area=an.width*an.height*(an.scale or 1)^2
		end
		local v=math.sqrt(weight*1/area)*4
		local vm=liftoff.max_speed(plane,0.2)*.9
		res= (v<vm) and v or vm
	end
	return res
end
local max_speed=function(plane,cons_modifier)
	--roughly estimates the speed of vehicle 
	--based on definitely incorrect formula
	local p=plane
	local consumption=liftoff.to_watt(p.consumption)*(cons_modifier or 1);
	return math.sqrt(p.effectivity*consumption/p.weight/p.friction)/770 
end

local multipliers={m=1e-3,k=1e3,M=1e6,G=1e9,T=1e12}
local to_watt=function(str)
	--converts power value from the string to number value 
	local _,_,num,char=string.find(str,"(%d+)(%a?)[Ww]")
	local m=multipliers[char] or 1
	return num*m
end
liftoff.to_watt=to_watt
liftoff.max_speed=max_speed
liftoff.lift_equation=lift_equation
liftoff.generate=generate
liftoff.import_planes=import_planes
